package presence

import (
	"log"
	"time"

	"github.com/coreos/go-etcd/etcd"
)

// EtcdClient is the etcd client interface used by this package.
type EtcdClient interface {
	AddChild(key string, value string, ttl uint64) (*etcd.Response, error)
	Create(key string, value string, ttl uint64) (*etcd.Response, error)
	CreateDir(key string, ttl uint64) (*etcd.Response, error)
	Update(key string, value string, ttl uint64) (*etcd.Response, error)
	Delete(key string, recursive bool) (*etcd.Response, error)
	Get(key string, sort, recursive bool) (*etcd.Response, error)
}

// A Client can read the list of active nodes written by Presence workers.
type Client struct {
	client EtcdClient
	path   string
}

func NewClient(client EtcdClient, path string) *Client {
	return &Client{
		client: client,
		path:   path,
	}
}

// Nodes returns a list of all the other running nodes.
func (c *Client) Nodes() ([]string, error) {
	resp, err := c.client.Get(c.path, false, true)
	if err != nil {
		return nil, err
	}
	var out []string
	for _, r := range resp.Node.Nodes {
		out = append(out, r.Value)
	}
	return out, nil
}

// StateFn is a function that returns the encoded state of the current node.
type StateFn func() string

// Presence maintains a list of active nodes on etcd, identified by a
// 'tag'. Nodes can report a small amount of information, periodically
// updated, that will be available to the other nodes.
type Presence struct {
	*Client
	nodeKey string
	stateFn StateFn
	ttl     uint64
	stop    chan struct{}
}

// New returns a new Presence worker that establishes node presence at
// the specified path.
func New(client EtcdClient, path string, stateFn StateFn, ttl uint64) *Presence {
	return &Presence{
		Client:  NewClient(client, path),
		ttl:     ttl,
		stateFn: stateFn,
		stop:    make(chan struct{}),
	}
}

// Stop the presence worker and release all resources.
func (p *Presence) Stop() {
	close(p.stop)
	p.client.Delete(p.nodeKey, false)
}

// Start the presence worker.
func (p *Presence) Start() error {
	// Create the presence directory if it does not exist. Ignore errors.
	p.client.CreateDir(p.path, 0)

	go p.run()
	return nil
}

func (p *Presence) run() {
	// Run 3 times faster than the TTL.
	ttl := time.Duration(p.ttl) * time.Second
	tick := time.NewTicker(ttl / 3)

	doinit := true
	lastUpdate := time.Now()

	for {
		select {
		case <-p.stop:
			return
		case t := <-tick.C:
			if doinit {
				// Create a new entry for the current node.
				resp, err := p.client.AddChild(p.path, p.stateFn(), p.ttl)
				if err != nil {
					log.Printf("could not create presence node: %v", err)
					continue
				}
				p.nodeKey = resp.Node.Key
				log.Printf("presence at %s", p.nodeKey)
				doinit = false
			} else {
				// Update current entry.
				if _, err := p.client.Update(p.nodeKey, p.stateFn(), p.ttl); err != nil {
					log.Printf("error updating presence: %v", err)
					// If we have received errors
					// for longer than a ttl,
					// create a new unique file.
					if t.Sub(lastUpdate) > ttl {
						doinit = true
					}
				}
			}
			lastUpdate = t
		}
	}
}
