package watcher

import (
	"sync"

	"github.com/coreos/go-etcd/etcd"
)

type DecodeFn func(string) interface{}

type hasClose interface {
	Close() error
}

type Cache struct {
	w        *Watcher
	decodeFn DecodeFn

	lock  sync.RWMutex
	index uint64
	data  map[string]interface{}
}

func defaultDecodeFn(value string) interface{} {
	return value
}

func NewCache(client EtcdClient, path string, decodeFn DecodeFn) *Cache {
	if decodeFn == nil {
		decodeFn = defaultDecodeFn
	}
	w := NewWatcher(client, path)
	c := &Cache{
		w:        w,
		decodeFn: decodeFn,
		data:     make(map[string]interface{}),
	}
	respCh := w.Start()
	go func() {
		for resp := range respCh {
			c.handleResponse(resp)
		}
	}()
	return c
}

func (c *Cache) Close() {
	c.w.Stop()
}

func flatten(nodes etcd.Nodes, n *etcd.Node) etcd.Nodes {
	if n.Dir {
		for _, nn := range n.Nodes {
			nodes = flatten(nodes, nn)
		}
	} else {
		nodes = append(nodes, n)
	}
	return nodes
}

func (c *Cache) handleResponse(resp *etcd.Response) {
	// Flatten out the node hierarchy.
	nodes := flatten(nil, resp.Node)

	c.lock.Lock()

	switch resp.Action {
	case "delete":
		for _, n := range nodes {
			if value, ok := c.data[n.Key]; ok {
				// If the object in the cache has a
				// Close method, invoke it in a
				// separate goroutine.
				if closer, ok := value.(hasClose); ok {
					go closer.Close()
				}
				delete(c.data, n.Key)
			}
		}

	case "set", "create", "update":
		for _, n := range nodes {
			c.data[n.Key] = c.decodeFn(n.Value)
		}
	}

	c.index = resp.EtcdIndex

	c.lock.Unlock()
}

func (c *Cache) Get(key string) (interface{}, bool) {
	c.lock.RLock()
	value, ok := c.data[key]
	c.lock.RUnlock()
	return value, ok
}

func (c *Cache) GetWithIndex(key string) (interface{}, uint64, bool) {
	c.lock.RLock()
	index := c.index
	value, ok := c.data[key]
	c.lock.RUnlock()
	return value, index, ok
}
