package watcher

import (
	"log"
	"time"

	"github.com/coreos/go-etcd/etcd"
)

// EtcdClient is the etcd client interface used by this package.
type EtcdClient interface {
	Create(key string, value string, ttl uint64) (*etcd.Response, error)
	Delete(key string, recursive bool) (*etcd.Response, error)
	Get(key string, sort, recursive bool) (*etcd.Response, error)
	CompareAndSwap(key string, value string, ttl uint64, prevValue string, prevIndex uint64) (*etcd.Response, error)
	Watch(prefix string, waitIndex uint64, recursive bool, receiver chan *etcd.Response, stop chan bool) (*etcd.Response, error)
}

// A Watcher will listen for changes on an etcd directory tree, and it
// will report them to the caller on the channel returned by Start().
type Watcher struct {
	client EtcdClient
	path   string
	stop   chan bool
	loaded chan struct{}
}

// NewWatcher returns a new Watcher on the specified path.
func NewWatcher(client EtcdClient, path string) *Watcher {
	return &Watcher{
		client: client,
		path:   path,
		stop:   make(chan bool),
		loaded: make(chan struct{}, 1),
	}
}

// Start the watcher. Changes will be sent on the returned channel.
// The Watcher will close the channel when Stop() is called.
func (w *Watcher) Start() <-chan *etcd.Response {
	c := make(chan *etcd.Response, 1)
	go w.watcher(c)
	return c
}

// WaitForInit waits until the initial load of the database data has
// succeeded.
func (w *Watcher) WaitForInit() {
	<-w.loaded
}

// Stop the Watcher and release all resources. Does not wait for the
// watcher goroutine to terminate.
func (w *Watcher) Stop() {
	close(w.stop)
}

func (w *Watcher) load(c chan<- *etcd.Response) uint64 {
	defer func() {
		// Do a nonblocking send on 'loaded' because load()
		// might be called more than once (for instance
		// whenever Watch returns with a 401 error).
		select {
		case w.loaded <- struct{}{}:
		default:
		}
	}()

	// Try to fetch the initial contents of the data below our
	// root. Repeat until we succeed, or we get a non-permanent
	// error from etcd.
	for {
		resp, err := w.client.Get(w.path, false, true)
		if err != nil {
			// TODO: check error code!
			if etcdErr, ok := err.(*etcd.EtcdError); ok {
				return etcdErr.Index
			}
			log.Printf("Get(%s) error: %v", w.path, err)
			if waitOrStop(200*time.Millisecond, w.stop) {
				return 0
			}
			continue
		}

		// Pretend it is a 'set' operation.
		resp.Action = "set"

		c <- resp

		return resp.EtcdIndex
	}
}

func (w *Watcher) watcher(c chan<- *etcd.Response) {
	defer close(c)

	// Avoid leaking the 'loaded' channel when we are stopped.
	defer close(w.loaded)

	index := w.load(c)

	for {
		resp, err := w.client.Watch(w.path, index+1, true, nil, w.stop)
		if err == etcd.ErrWatchStoppedByUser {
			return
		}
		if err != nil {
			if etcdErr, ok := err.(*etcd.EtcdError); ok {
				// If the index passed to Watch falls
				// behind the current log position by
				// more than 1000 entries (which is
				// bound to happen regularly in
				// presence of a sustained write
				// rate), etcd will return an error
				// 401 ("index out of date"). The safe
				// thing to do is just to run a full
				// load again, and resume watching
				// from the current EtcdIndex.
				if etcdErr.ErrorCode == 401 {
					index = w.load(c)
					continue
				}
				index = etcdErr.Index
			}
			log.Printf("Watch(%s) error: %v", w.path, err)
			time.Sleep(200 * time.Millisecond)
			continue
		}

		index = resp.EtcdIndex

		c <- resp
	}
}

func waitOrStop(t time.Duration, stop <-chan bool) bool {
	a := time.After(t)
	select {
	case <-a:
	case <-stop:
		return true
	}
	return false
}
